# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import multiprocessing
import sys

import lightning as pl
import torch

from clapper.logging import setup
from torch.utils.data import DataLoader, WeightedRandomSampler
from torchvision import transforms

from .dataset import get_samples_weights

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


class BaseDataModule(pl.LightningDataModule):
    def __init__(
        self,
        batch_size=1,
        batch_chunk_count=1,
        drop_incomplete_batch=False,
        parallel=-1,
    ):
        super().__init__()

        self.batch_size = batch_size
        self.batch_chunk_count = batch_chunk_count

        self.train_dataset = None
        self.validation_dataset = None
        self.extra_validation_datasets = None

        self._raw_data_transforms = []
        self._model_transforms = []
        self._augmentation_transforms = []

        self.drop_incomplete_batch = drop_incomplete_batch
        self.pin_memory = (
            torch.cuda.is_available()
        )  # should only be true if GPU available and using it

        self.parallel = parallel

    def setup(self, stage: str):
        # Implemented by user
        # Must define self.train_dataset, self.validation_dataset, self.extra_validation_datasets and self.predict_dataset
        raise NotImplementedError

    def train_dataloader(self):
        train_samples_weights = get_samples_weights(self.train_dataset)

        multiproc_kwargs = self._setup_multiproc(self.parallel)
        train_sampler = WeightedRandomSampler(
            train_samples_weights, len(train_samples_weights), replacement=True
        )

        return DataLoader(
            self.train_dataset,
            batch_size=self._compute_chunk_size(self.batch_size),
            drop_last=self.drop_incomplete_batch,
            pin_memory=self.pin_memory,
            sampler=train_sampler,
            **multiproc_kwargs,
        )

    def val_dataloader(self):
        loaders_dict = {}

        multiproc_kwargs = self._setup_multiproc(self.parallel)

        val_loader = DataLoader(
            dataset=self.validation_dataset,
            batch_size=self._compute_chunk_size(self.batch_size),
            shuffle=False,
            drop_last=False,
            pin_memory=self.pin_memory,
            **multiproc_kwargs,
        )

        loaders_dict["validation_loader"] = val_loader

        if self.extra_validation_datasets is not None:
            for set_idx, extra_set in enumerate(self.extra_validation_datasets):
                extra_val_loader = DataLoader(
                    dataset=extra_set,
                    batch_size=self._compute_chunk_size(self.batch_size),
                    shuffle=False,
                    drop_last=False,
                    pin_memory=self.pin_memory,
                    **multiproc_kwargs,
                )

                loaders_dict[
                    f"extra_validation_loader{set_idx}"
                ] = extra_val_loader

        return loaders_dict

    def predict_dataloader(self):
        loaders_dict = {}

        loaders_dict["train_dataloader"] = self.train_dataloader()
        for k, v in self.val_dataloader().items():
            loaders_dict[k] = v

        return loaders_dict

    def _compute_chunk_size(self, batch_size):
        batch_chunk_size = batch_size
        if batch_size % self.batch_chunk_count != 0:
            # batch_size must be divisible by batch_chunk_count.
            raise RuntimeError(
                f"--batch-size ({batch_size}) must be divisible by "
                f"--batch-chunk-size ({self.batch_chunk_count})."
            )
        else:
            batch_chunk_size = batch_size // self.batch_chunk_count

        return batch_chunk_size

    def _setup_multiproc(self, parallel):
        multiproc_kwargs = dict()
        if parallel < 0:
            multiproc_kwargs["num_workers"] = 0
        else:
            multiproc_kwargs["num_workers"] = (
                parallel or multiprocessing.cpu_count()
            )

        if multiproc_kwargs["num_workers"] > 0 and sys.platform == "darwin":
            multiproc_kwargs[
                "multiprocessing_context"
            ] = multiprocessing.get_context("spawn")

        return multiproc_kwargs

    def _build_transforms(self, is_train):
        all_transforms = self.raw_data_transforms + self.model_transforms

        if is_train:
            all_transforms = all_transforms + self.augmentation_transforms

        # all_transforms.append(transforms.ToTensor())

        return transforms.Compose(all_transforms)

    @property
    def raw_data_transforms(self):
        return self._raw_data_transforms

    @raw_data_transforms.setter
    def raw_data_transforms(self, transforms):
        self._raw_data_transforms = transforms

    @property
    def model_transforms(self):
        return self._model_transforms

    @model_transforms.setter
    def model_transforms(self, transforms):
        self._model_transforms = transforms

    @property
    def augmentation_transforms(self):
        return self._augmentation_transforms

    @augmentation_transforms.setter
    def augmentation_transforms(self, transforms):
        self._augmentation_transforms = transforms


def get_dataset_from_module(module, stage, **module_args):
    """Instantiates a DataModule and retrieves the corresponding dataset.

    Useful when combining multiple datasets.
    """
    module_instance = module(**module_args)
    module_instance.prepare_data()
    module_instance.setup(stage=stage)
    dataset = module_instance.dataset

    return dataset
