# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import json
import logging
import os
import pathlib

import torch

from tqdm import tqdm

logger = logging.getLogger(__name__)


class JSONProtocol:
    """Generic multi-protocol/subset filelist dataset that yields samples.

    To create a new dataset, you need to provide one or more JSON formatted
    filelists (one per protocol) with the following contents:

    .. code-block:: json

       {
           "subset1": [
               [
                   "value1",
                   "value2",
                   "value3"
               ],
               [
                   "value4",
                   "value5",
                   "value6"
               ]
           ],
           "subset2": [
           ]
       }

    Your dataset many contain any number of subsets, but all sample entries
    must contain the same number of fields.


    Parameters
    ----------

    protocols : list, dict
        Paths to one or more JSON formatted files containing the various
        protocols to be recognized by this dataset, or a dictionary, mapping
        protocol names to paths (or opened file objects) of CSV files.
        Internally, we save a dictionary where keys default to the basename of
        paths (list input).

    fieldnames : list, tuple
        An iterable over the field names (strings) to assign to each entry in
        the JSON file.  It should have as many items as fields in each entry of
        the JSON file.

    loader : object
        A function that receives as input, a context dictionary (with at least
        a "protocol" and "subset" keys indicating which protocol and subset are
        being served), and a dictionary with ``{fieldname: value}`` entries,
        and returns an object with at least 2 attributes:

        * ``key``: which must be a unique string for every sample across
          subsets in a protocol, and
        * ``data``: which contains the data associated witht this sample
    """

    def __init__(self, protocols, fieldnames):
        if isinstance(protocols, dict):
            self._protocols = protocols
        else:
            self._protocols = {
                os.path.basename(
                    str(k).replace("".join(pathlib.Path(k).suffixes), "")
                ): k
                for k in protocols
            }
        self.fieldnames = fieldnames

    def check(self, limit=0):
        """For each protocol, check if all data can be correctly accessed.

        This function assumes each sample has a ``data`` and a ``key``
        attribute.  The ``key`` attribute should be a string, or representable
        as such.


        Parameters
        ----------

        limit : int
            Maximum number of samples to check (in each protocol/subset
            combination) in this dataset.  If set to zero, then check
            everything.


        Returns
        -------

        errors : int
            Number of errors found
        """
        logger.info("Checking dataset...")
        errors = 0
        for proto in self._protocols:
            logger.info(f"Checking protocol '{proto}'...")
            for name, samples in self.subsets(proto).items():
                logger.info(f"Checking subset '{name}'...")
                if limit:
                    logger.info(f"Checking at most first '{limit}' samples...")
                    samples = samples[:limit]
                for pos, sample in enumerate(samples):
                    try:
                        sample.data  # may trigger data loading
                        logger.info(f"{sample.key}: OK")
                    except Exception as e:
                        logger.error(
                            f"Found error loading entry {pos} in subset {name} "
                            f"of protocol {proto} from file "
                            f"'{self._protocols[proto]}': {e}"
                        )
                        errors += 1
        return errors

    def subsets(self, protocol):
        """Returns all subsets in a protocol.

        This method will load JSON information for a given protocol and return
        all subsets of the given protocol after converting each entry through
        the loader function.

        Parameters
        ----------

        protocol : str
            Name of the protocol data to load


        Returns
        -------

        subsets : dict
            A dictionary mapping subset names to lists of objects (respecting
            the ``key``, ``data`` interface).
        """
        fileobj = self._protocols[protocol]
        if isinstance(fileobj, (str, bytes, pathlib.Path)):
            if str(fileobj).endswith(".bz2"):
                import bz2

                with bz2.open(self._protocols[protocol]) as f:
                    data = json.load(f)
            else:
                with open(self._protocols[protocol]) as f:
                    data = json.load(f)
        else:
            data = json.load(fileobj)
            fileobj.seek(0)

        retval = {}
        for subset, samples in data.items():
            logger.info(f"Loading subset {subset} samples.")

            retval[subset] = [
                dict(zip(self.fieldnames, k))
                for n, k in enumerate(tqdm(samples))
            ]

        return retval


class CSVDataset:
    """Generic multi-subset filelist dataset that yields samples.

    To create a new dataset, you only need to provide a CSV formatted filelist
    using any separator (e.g. comma, space, semi-colon) with the following
    information:

    .. code-block:: text

       value1,value2,value3
       value4,value5,value6
       ...

    Notice that all rows must have the same number of entries.

    Parameters
    ----------

    subsets : list, dict
        Paths to one or more CSV formatted files containing the various subsets
        to be recognized by this dataset, or a dictionary, mapping subset names
        to paths (or opened file objects) of CSV files.  Internally, we save a
        dictionary where keys default to the basename of paths (list input).

    fieldnames : list, tuple
        An iterable over the field names (strings) to assign to each column in
        the CSV file.  It should have as many items as fields in each row of
        the CSV file(s).

    loader : object
        A function that receives as input, a context dictionary (with, at
        least, a "subset" key indicating which subset is being served), and a
        dictionary with ``{key: path}`` entries, and returns a dictionary with
        the loaded data.
    """

    def __init__(self, subsets, fieldnames, loader):
        if isinstance(subsets, dict):
            self._subsets = subsets
        else:
            self._subsets = {
                os.path.basename(
                    str(k).replace("".join(pathlib.Path(k).suffixes), "")
                ): k
                for k in subsets
            }
        self.fieldnames = fieldnames
        self._loader = loader

    def check(self, limit=0):
        """For each subset, check if all data can be correctly accessed.

        This function assumes each sample has a ``data`` and a ``key``
        attribute.  The ``key`` attribute should be a string, or representable
        as such.


        Parameters
        ----------

        limit : int
            Maximum number of samples to check (in each protocol/subset
            combination) in this dataset.  If set to zero, then check
            everything.


        Returns
        -------

        errors : int
            Number of errors found
        """
        logger.info("Checking dataset...")
        errors = 0
        for name in self._subsets.keys():
            logger.info(f"Checking subset '{name}'...")
            samples = self.samples(name)
            if limit:
                logger.info(f"Checking at most first '{limit}' samples...")
                samples = samples[:limit]
            for pos, sample in enumerate(samples):
                try:
                    sample.data  # may trigger data loading
                    logger.info(f"{sample.key}: OK")
                except Exception as e:
                    logger.error(
                        f"Found error loading entry {pos} in subset {name} "
                        f"from file '{self._subsets[name]}': {e}"
                    )
                    errors += 1
        return errors

    def subsets(self):
        """Returns all available subsets at once.

        Returns
        -------

        subsets : dict
            A dictionary mapping subset names to lists of objects (respecting
            the ``key``, ``data`` interface).
        """
        return {k: self.samples(k) for k in self._subsets.keys()}

    def samples(self, subset):
        """Returns all samples in a subset.

        This method will load CSV information for a given subset and return
        all samples of the given subset after passing each entry through the
        loading function.


        Parameters
        ----------

        subset : str
            Name of the subset data to load


        Returns
        -------

        subset : list
            A lists of objects (respecting the ``key``, ``data`` interface).
        """
        fileobj = self._subsets[subset]
        if isinstance(fileobj, (str, bytes, pathlib.Path)):
            with open(self._subsets[subset], newline="") as f:
                cf = csv.reader(f)
                samples = [k for k in cf]
        else:
            cf = csv.reader(fileobj)
            samples = [k for k in cf]
            fileobj.seek(0)

        return [
            self._loader(
                dict(subset=subset, order=n), dict(zip(self.fieldnames, k))
            )
            for n, k in enumerate(samples)
        ]


class CachedDataset(torch.utils.data.Dataset):
    def __init__(
        self, json_protocol, protocol, subset, raw_data_loader, transforms
    ):
        self.json_protocol = json_protocol
        self.subset = subset
        self.raw_data_loader = raw_data_loader
        self.transforms = transforms

        self._samples = json_protocol.subsets(protocol)[self.subset]
        # Dict entry with relative path to files, used during prediction
        for s in self._samples:
            s["name"] = s["data"]

        logger.info(f"Caching {self.subset} samples")
        for sample in tqdm(self._samples):
            sample["data"] = self.transforms(
                self.raw_data_loader(sample["data"])
            )

    def __getitem__(self, idx):
        return self._samples[idx]

    def __len__(self):
        return len(self._samples)


class RuntimeDataset(torch.utils.data.Dataset):
    def __init__(
        self, json_protocol, protocol, subset, raw_data_loader, transforms
    ):
        self.json_protocol = json_protocol
        self.subset = subset
        self.raw_data_loader = raw_data_loader
        self.transforms = transforms

        self._samples = json_protocol.subsets(protocol)[self.subset]
        # Dict entry with relative path to files
        for s in self._samples:
            s["name"] = s["data"]

    def __getitem__(self, idx):
        sample = self._samples[idx].copy()
        sample["data"] = self.transforms(self.raw_data_loader(sample["data"]))
        return sample

    def __len__(self):
        return len(self._samples)


class TBDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        json_protocol,
        protocol,
        subset,
        raw_data_loader,
        transforms,
        cache_samples=False,
    ):
        self.json_protocol = json_protocol
        self.subset = subset
        self.raw_data_loader = raw_data_loader
        self.transforms = transforms

        self.cache_samples = cache_samples

        self._samples = json_protocol.subsets(protocol)[self.subset]

        # Dict entry with relative path to files
        for s in self._samples:
            s["name"] = s["data"]

        if self.cache_samples:
            logger.info(f"Caching {self.subset} samples")
            for sample in tqdm(self._samples):
                sample["data"] = self.transforms(
                    self.raw_data_loader(sample["data"])
                )

    def __getitem__(self, idx):
        if self.cache_samples:
            return self._samples[idx]
        else:
            sample = self._samples[idx].copy()
            sample["data"] = self.transforms(
                self.raw_data_loader(sample["data"])
            )
            return sample

    def __len__(self):
        return len(self._samples)


def get_samples_weights(dataset):
    """Compute the weights of all the samples of the dataset to balance it
    using the sampler of the dataloader.

    This function takes as input a :py:class:`torch.utils.data.dataset.Dataset`
    and computes the weights to balance each class in the dataset and the
    datasets themselves if we have a ConcatDataset.


    Parameters
    ----------

    dataset : torch.utils.data.dataset.Dataset
        An instance of torch.utils.data.dataset.Dataset
        ConcatDataset are supported


    Returns
    -------

    samples_weights : :py:class:`torch.Tensor`
        the weights for all the samples in the dataset given as input
    """
    samples_weights = []

    if isinstance(dataset, torch.utils.data.ConcatDataset):
        for ds in dataset.datasets:
            # Weighting only for binary labels
            if isinstance(ds._samples[0]["label"], int):
                # Groundtruth
                targets = []
                for s in ds._samples:
                    targets.append(s["label"])
                targets = torch.tensor(targets)

                # Count number of samples per class
                class_sample_count = torch.tensor(
                    [
                        (targets == t).sum()
                        for t in torch.unique(targets, sorted=True)
                    ]
                )

                weight = 1.0 / class_sample_count.float()

                samples_weights.append(
                    torch.tensor([weight[t] for t in targets])
                )

            else:
                # We only weight to sample equally from each dataset
                samples_weights.append(torch.full((len(ds),), 1.0 / len(ds)))

        # Concatenate sample weights from all the datasets
        samples_weights = torch.cat(samples_weights)

    else:
        # Weighting only for binary labels
        if isinstance(dataset._samples[0]["label"], int):
            # Groundtruth
            targets = []
            for s in dataset._samples:
                targets.append(s["label"])
            targets = torch.tensor(targets)

            # Count number of samples per class
            class_sample_count = torch.tensor(
                [
                    (targets == t).sum()
                    for t in torch.unique(targets, sorted=True)
                ]
            )

            weight = 1.0 / class_sample_count.float()

            samples_weights = torch.tensor([weight[t] for t in targets])

        else:
            # Equal weights for non-binary labels
            samples_weights = torch.ones(len(dataset._samples))

    return samples_weights


def get_positive_weights(dataset):
    """Compute the positive weights of each class of the dataset to balance the
    BCEWithLogitsLoss criterion.

    This function takes as input a :py:class:`torch.utils.data.dataset.Dataset`
    and computes the positive weights of each class to use them to have
    a balanced loss.


    Parameters
    ----------

    dataset : torch.utils.data.dataset.Dataset
        An instance of torch.utils.data.dataset.Dataset
        ConcatDataset are supported


    Returns
    -------

    positive_weights : :py:class:`torch.Tensor`
        the positive weight of each class in the dataset given as input
    """
    targets = []

    if isinstance(dataset, torch.utils.data.ConcatDataset):
        for ds in dataset.datasets:
            for s in ds._samples:
                targets.append(s["label"])

    else:
        for s in dataset._samples:
            targets.append(s["label"])

    targets = torch.tensor(targets)

    # Binary labels
    if len(list(targets.shape)) == 1:
        class_sample_count = [
            float((targets == t).sum().item())
            for t in torch.unique(targets, sorted=True)
        ]

        # Divide negatives by positives
        positive_weights = torch.tensor(
            [class_sample_count[0] / class_sample_count[1]]
        ).reshape(-1)

    # Multiclass labels
    else:
        class_sample_count = torch.sum(targets, dim=0)
        negative_class_sample_count = (
            torch.full((targets.size()[1],), float(targets.size()[0]))
            - class_sample_count
        )

        positive_weights = negative_class_sample_count / (
            class_sample_count + negative_class_sample_count
        )

    return positive_weights


def reweight_BCEWithLogitsLoss(datamodule, model, criterion, criterion_valid):
    from torch.nn import BCEWithLogitsLoss

    datamodule.prepare_data()
    datamodule.setup(stage="fit")

    train_dataset = datamodule.train_dataset
    validation_dataset = datamodule.validation_dataset

    # Redefine a weighted criterion if possible
    if isinstance(criterion, torch.nn.BCEWithLogitsLoss):
        positive_weights = get_positive_weights(train_dataset)
        model.hparams.criterion = BCEWithLogitsLoss(pos_weight=positive_weights)
    else:
        logger.warning("Weighted criterion not supported")

    if validation_dataset is not None:
        # Redefine a weighted valid criterion if possible
        if (
            isinstance(criterion_valid, torch.nn.BCEWithLogitsLoss)
            or criterion_valid is None
        ):
            positive_weights = get_positive_weights(validation_dataset)
            model.hparams.criterion_valid = BCEWithLogitsLoss(
                pos_weight=positive_weights
            )
        else:
            logger.warning("Weighted valid criterion not supported")


def normalize_data(normalization, model, datamodule):
    from torch.utils.data import DataLoader

    datamodule.prepare_data()
    datamodule.setup(stage="fit")

    train_dataset = datamodule.train_dataset

    # Create z-normalization model layer if needed
    if normalization == "imagenet":
        model.normalizer.set_mean_std(
            [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
        )
        logger.info("Z-normalization with ImageNet mean and std")
    elif normalization == "current":
        # Compute mean/std of current train subset
        temp_dl = DataLoader(
            dataset=train_dataset, batch_size=len(train_dataset)
        )

        data = next(iter(temp_dl))
        mean = data[1].mean(dim=[0, 2, 3])
        std = data[1].std(dim=[0, 2, 3])

        model.normalizer.set_mean_std(mean, std)

        # Format mean and std for logging
        mean = str(
            [
                round(x, 3)
                for x in ((mean * 10**3).round() / (10**3)).tolist()
            ]
        )
        std = str(
            [
                round(x, 3)
                for x in ((std * 10**3).round() / (10**3)).tolist()
            ]
        )
        logger.info(f"Z-normalization with mean {mean} and std {std}")
