# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shenzhen dataset for TB detection (default protocol)

* Split reference: first 64% of TB and healthy CXR for "train" 16% for
* "validation", 20% for "test"
* This configuration resolution: 512 x 512 (default)
* See :py:mod:`ptbench.data.shenzhen` for dataset details
"""

from clapper.logging import setup
from torchvision import transforms

from ..base_datamodule import BaseDataModule
from ..dataset import JSONProtocol, TBDataset
from ..shenzhen import _protocols, _raw_data_loader
from ..transforms import ElasticDeformation, RemoveBlackBorders

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


class DefaultModule(BaseDataModule):
    def __init__(
        self,
        batch_size=1,
        batch_chunk_count=1,
        drop_incomplete_batch=False,
        cache_samples=False,
        parallel=-1,
    ):
        super().__init__(
            batch_size=batch_size,
            drop_incomplete_batch=drop_incomplete_batch,
            batch_chunk_count=batch_chunk_count,
            parallel=parallel,
        )

        self._cache_samples = cache_samples
        self._has_setup_fit = False
        self._has_setup_predict = False
        self._protocol = "default"

        self.raw_data_transforms = [
            RemoveBlackBorders(),
            transforms.Resize(512),
            transforms.CenterCrop(512),
            transforms.ToTensor(),
        ]

        self.model_transforms = []

        self.augmentation_transforms = [ElasticDeformation(p=0.8)]

    def setup(self, stage: str):
        json_protocol = JSONProtocol(
            protocols=_protocols,
            fieldnames=("data", "label"),
        )

        if not self._has_setup_fit and stage == "fit":
            self.train_dataset = TBDataset(
                json_protocol,
                self._protocol,
                "train",
                _raw_data_loader,
                self._build_transforms(is_train=True),
                cache_samples=self._cache_samples,
            )
            self.validation_dataset = TBDataset(
                json_protocol,
                self._protocol,
                "validation",
                _raw_data_loader,
                self._build_transforms(is_train=False),
                cache_samples=self._cache_samples,
            )

            self._has_setup_fit = True

        if not self._has_setup_predict and stage == "predict":
            self.train_dataset = TBDataset(
                json_protocol,
                self._protocol,
                "train",
                _raw_data_loader,
                self._build_transforms(is_train=False),
                cache_samples=self._cache_samples,
            )
            self.validation_dataset = TBDataset(
                json_protocol,
                self._protocol,
                "validation",
                _raw_data_loader,
                self._build_transforms(is_train=False),
                cache_samples=self._cache_samples,
            )

            self._has_setup_predict = True


datamodule = DefaultModule
