# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shenzhen dataset for TB detection (cross validation fold 0, RGB)

* Split reference: first 80% of TB and healthy CXR for "train", rest for "test"
* This configuration resolution: 512 x 512 (default)
* See :py:mod:`ptbench.data.shenzhen` for dataset details
"""

from clapper.logging import setup

from ....data import return_subsets
from ....data.base_datamodule import BaseDataModule
from ....data.dataset import JSONProtocol
from ....data.shenzhen import _cached_loader, _delayed_loader, _protocols

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


class DefaultModule(BaseDataModule):
    def __init__(
        self,
        train_batch_size=1,
        predict_batch_size=1,
        drop_incomplete_batch=False,
        cache_samples=False,
        multiproc_kwargs=None,
        data_transforms=[],
        model_transforms=[],
        train_transforms=[],
    ):
        super().__init__(
            train_batch_size=train_batch_size,
            predict_batch_size=predict_batch_size,
            drop_incomplete_batch=drop_incomplete_batch,
            multiproc_kwargs=multiproc_kwargs,
        )

        self.cache_samples = cache_samples
        self.has_setup_fit = False

        self.data_transforms = data_transforms
        self.model_transforms = model_transforms
        self.train_transforms = train_transforms

        """[
            transforms.ToPILImage(),
            transforms.Lambda(lambda x: x.convert("RGB")),
            transforms.ToTensor(),
        ]"""

    def setup(self, stage: str):
        if self.cache_samples:
            logger.info(
                "Argument cache_samples set to True. Samples will be loaded in memory."
            )
            samples_loader = _cached_loader
        else:
            logger.info(
                "Argument cache_samples set to False. Samples will be loaded at runtime."
            )
            samples_loader = _delayed_loader

        self.json_protocol = JSONProtocol(
            protocols=_protocols,
            fieldnames=("data", "label"),
            loader=samples_loader,
            post_transforms=self.post_transforms,
        )

        if not self.has_setup_fit and stage == "fit":
            (
                self.train_dataset,
                self.validation_dataset,
                self.extra_validation_datasets,
            ) = return_subsets(self.json_protocol, "default", stage)
            self.has_setup_fit = True


datamodule = DefaultModule
